package org.netkernel.layer0.util;

import java.io.*;
import java.net.InetAddress;
import java.net.URL;
import java.net.URLConnection;
import java.net.UnknownHostException;
import java.nio.charset.Charset;
import java.text.DecimalFormat;
import java.util.*;

import org.netkernel.container.ILogger;
import org.netkernel.container.config.IConfiguration;
import org.netkernel.urii.INetKernelException;
import org.netkernel.util.Utils;


public class PropertyConfiguration implements IConfiguration
{
	private final Properties mProperties = new Properties();
	private final Map mTypedCache = new HashMap();
	private final ILogger mLogger;
	private final URL mPropertiesResource;
	private final long mTimestamp;
	private final PairList mAdditionalProperties = new PairList(4);
	
	public PropertyConfiguration(URL aPropertiesResource, ILogger aLogger) throws IOException
	{	InputStream is = aPropertiesResource.openStream();
		//This is not backwards compatible with Java5
		InputStreamReader  isr=new InputStreamReader(is, Charset.forName("UTF-8"));
		mProperties.load(isr);
		mLogger=aLogger;
		
		if (aPropertiesResource.getProtocol().equals("file"))
		{	mPropertiesResource=aPropertiesResource;
			mTimestamp=mPropertiesResource.openConnection().getLastModified();
		}
		else
		{	mPropertiesResource=null;
			mTimestamp=0;
		}
		String instanceIdentifier=this.getString("netkernel.instance.identifier","");
		if (instanceIdentifier.length()==0)
		{	try
			{	InetAddress localMachine = InetAddress.getLocalHost();
				this.setProperty("netkernel.instance.identifier", localMachine.getHostName());
			}
			catch (UnknownHostException e)
			{	this.setProperty("netkernel.instance.identifier", "unknown");
			}
		}
		try
		{	this.setProperty("netkernel.corePerformance",new DecimalFormat("0.00").format(Layer0Factory.getCorePerformanceUnit()));
		}
		catch (Exception e)	 {;}
		this.setProperty("netkernel.edition",Layer0Factory.getEdition());
	}
	
	/** Construct a new PropertyConfiguration based on a previous instance.
	 * resource is re-read but additional added values are preserved
	 **/
	public PropertyConfiguration(PropertyConfiguration aOther) throws IOException
	{	this(aOther.mPropertiesResource,aOther.mLogger);
		for (int i=0; i<aOther.mAdditionalProperties.size(); i++)
		{	String key=(String)aOther.mAdditionalProperties.getValue1(i);
			String value=(String)aOther.mAdditionalProperties.getValue2(i);
			this.setProperty(key, value);
		}
	}
	
	public boolean isExpired()
	{	boolean result=mPropertiesResource!=null;
		if (result)
		{	URLConnection con=null;
			try
			{	con=mPropertiesResource.openConnection();
				result = mTimestamp!=con.getLastModified();
			}
			catch (IOException e)
			{
			}
			finally
			{	if(con!=null)
				{	try
					{	//Fix JVM bug that leaves a file handle open after each getLastModified 
						con.getInputStream().close();
					}
					catch(IOException e2)
					{
					}
				}
			}
		}
		return result;
	}
	
	public URL getResource()
	{	return mPropertiesResource;
	}
	
	
	public String getString(String aKey) throws INetKernelException
	{	String result=mProperties.getProperty(aKey);
		if (result==null)
		{	throw Utils.createFormattedException("EX_UNDEFINED_CONFIG",aKey,mLogger);
		}
		return result;
	}

	public String getString(String aKey, String aDefault)
	{	String result=mProperties.getProperty(aKey);
		if (result==null)
		{	result=aDefault;
			mLogger.log(ILogger.WARNING,this,"MSG_USING_CONFIG_DEFAULT",aKey,aDefault);
		}
		return result;
	}

	public int getInt(String aKey) throws INetKernelException
	{	int result;
		Object typed = mTypedCache.get(aKey);
		if (typed instanceof Integer)
		{	result = ((Integer)typed).intValue();
		}
		else
		{	String unparsed = getString(aKey);
			try
			{ result = Integer.parseInt(unparsed);
				mTypedCache.put(aKey,new Integer(result));
			}
			catch (NumberFormatException e)
			{	throw Utils.createFormattedException("EX_MALFORMED_CONFIG",aKey,mLogger);
			}
		}
		return result;
	}

	public int getInt(String aKey, int aDefault)
	{	int result;
		Object typed = mTypedCache.get(aKey);
		if (typed instanceof Integer)
		{	result = ((Integer)typed).intValue();
		}
		else
		{	String unparsed = mProperties.getProperty(aKey);
			if (unparsed==null)
			{	result=aDefault;
				mLogger.log(ILogger.INFO,this,"MSG_USING_CONFIG_DEFAULT",aKey,Integer.toString(aDefault));
			}
			else
			{	try
				{ result = Integer.parseInt(unparsed);
				}
				catch (NumberFormatException e)
				{	result=aDefault;
				mLogger.log(ILogger.WARNING,this,"MSG_BAD_CONFIG_DEFAULT",aKey,Integer.toString(aDefault));
				}
			}
			mTypedCache.put(aKey,new Integer(result));
		}
		return result;
	}

	public long getLong(String aKey) throws INetKernelException
	{
		long result;
		Object typed = mTypedCache.get(aKey);
		if (typed instanceof Long)
		{	result = ((Long)typed).longValue();
		}
		else
		{	String unparsed = getString(aKey);
			try
			{ result = Long.parseLong(unparsed);
				mTypedCache.put(aKey,new Long(result));
			}
			catch (NumberFormatException e)
			{	throw Utils.createFormattedException("EX_MALFORMED_CONFIG",aKey,mLogger);
			}
		}
		return result;
	}

	public long getLong(String aKey, long aDefault)
	{
		long result;
		Object typed = mTypedCache.get(aKey);
		if (typed instanceof Long)
		{	result = ((Long)typed).longValue();
		}
		else
		{	String unparsed = mProperties.getProperty(aKey);
			if (unparsed==null)
			{	result=aDefault;
				mLogger.log(ILogger.INFO,this,"MSG_USING_CONFIG_DEFAULT",aKey,Long.toString(aDefault));
			}
			else
			{	try
				{ result = Long.parseLong(unparsed);
				}
				catch (NumberFormatException e)
				{	result=aDefault;
					mLogger.log(ILogger.WARNING,this,"MSG_BAD_CONFIG_DEFAULT",aKey,Long.toString(aDefault));
				}
			}
			mTypedCache.put(aKey,new Long(result));
		}
		return result;
	}

	public boolean getBoolean(String aKey) throws INetKernelException
	{
		boolean result;
		Object typed = mTypedCache.get(aKey);
		if (typed instanceof Boolean)
		{	result = ((Boolean)typed).booleanValue();
		}
		else
		{	String unparsed = getString(aKey);
			try
			{ result = Boolean.parseBoolean(unparsed);
				mTypedCache.put(aKey,new Boolean(result));
			}
			catch (NumberFormatException e)
			{	throw Utils.createFormattedException("EX_MALFORMED_CONFIG",aKey,mLogger);
			}
		}
		return result;
	}

	public boolean getBoolean(String aKey, boolean aDefault)
	{
		boolean result;
		Object typed = mTypedCache.get(aKey);
		if (typed instanceof Boolean)
		{	result = ((Boolean)typed).booleanValue();
		}
		else
		{	String unparsed = mProperties.getProperty(aKey);
			if (unparsed==null)
			{	result=aDefault;
				mLogger.log(ILogger.INFO,this,"MSG_USING_CONFIG_DEFAULT",aKey,Boolean.toString(aDefault));
			}
			else
			{	try
				{ result = Boolean.parseBoolean(unparsed);
				}
				catch (NumberFormatException e)
				{	result=aDefault;
				mLogger.log(ILogger.WARNING,this,"MSG_BAD_CONFIG_DEFAULT",aKey,Boolean.toString(aDefault));
				}
			}
			mTypedCache.put(aKey,new Boolean(result));
		}
		return result;
	}

	public Iterator getKeys()
	{	return mProperties.keySet().iterator();
	}
	
	public boolean containsKey(String key)
	{	return mProperties.containsKey(key);
	}
	
	public void setProperty(String key, String value)
	{	
		if (!mProperties.containsKey(key) && !mAdditionalProperties.containsValue1(key))
		{	mAdditionalProperties.put(key, value);
		}
		mProperties.setProperty(key, value);
		mTypedCache.clear();
	}
}
